from langgraph.graph import StateGraph, END, START

from src.utils.visualize import visualize_graph
from src.utils.state import State
from src.utils.nodes import call_chat_model, should_run

builder = StateGraph(State)

# Nodes
builder.add_node("call_chat_model", call_chat_model)

# Add conditional edge
builder.add_conditional_edges(START, should_run)
# Connect nodes to the end
builder.add_edge("call_chat_model", END)
graph = builder.compile()
# graph.debug = True
visualize_graph(graph, "should_run_end")


import os
from typing import Annotated, TypedDict

from langgraph.graph import StateGraph, add_messages, START, END
from langchain_core.messages import AnyMessage, SystemMessage

from src.utils.llm import LLMWrapper, ModelName
from src.utils.tools import tool_node

class State(TypedDict):
    system: str
    messages: Annotated[list[AnyMessage], add_messages]
    tools: list
    thread_id: int


def should_continue(state: State):
    messages = state["messages"]
    last_message = messages[-1]
    if last_message.tool_calls:
        return "actions"
    return END


def call_model(state: State):
    messages = state["messages"]
    # Get tools from state that match available tool names
    selected_tools = []
    if "tools" in state:
        from src.utils.tools import tools
        selected_tools = [tool for tool in tools if tool.name in state["tools"]]
    
    llm = LLMWrapper(
        model_name=ModelName.ANTHROPIC,
        api_key=os.getenv("ANTHROPIC_API_KEY"), 
        tools=selected_tools
    )
    if not isinstance(messages[0], SystemMessage):
        messages.insert(0, SystemMessage(content=state.get("system", "You are a helpful assistant.")))
    response = llm.invoke(messages)
    return {"messages": [response]}


builder = StateGraph(State)

# Define the two nodes we will cycle between
builder.add_node("agent", call_model)
builder.add_node("actions", tool_node)

builder.add_edge(START, "agent")
builder.add_conditional_edges("agent", should_continue, ["actions", END])
builder.add_edge("actions", "agent")


# https://python.langchain.com/docs/versions/migrating_memory/chat_history/
from typing_extensions import Annotated, TypedDict
from langgraph.graph.message import add_messages
from langchain_core.messages import BaseMessage
from langchain_core.chat_history import InMemoryChatMessageHistory
from langchain_core.runnables import RunnableConfig


class State(TypedDict):
    messages: Annotated[list, add_messages]
    session_id: str
    enabled: bool
    
class StateManager:
    def __init__(self, config: RunnableConfig):
        self.config = config
        self.session_id = config["configurable"]["session_id"]
        
    def get_chat_history(self, store: dict[str, InMemoryChatMessageHistory]) -> InMemoryChatMessageHistory:
        chat_history = store.get(self.session_id)
        if chat_history is None:
            chat_history = InMemoryChatMessageHistory()
            store[self.session_id] = chat_history
        return chat_history
    
    def collect_messages(self, store: dict[str, InMemoryChatMessageHistory], state: State) -> list[BaseMessage]:
        return list(self.get_chat_history(store).messages) + state["messages"]

    def add_messages(self, store: dict[str, InMemoryChatMessageHistory], messages: list[BaseMessage]):
        self.get_chat_history(store).add_messages(messages)



from enum import Enum

ROOT_PATH = './sandbox/system' 

class SystemPaths(Enum):
    DEFAULT = f"{ROOT_PATH}/default.txt"
    PIRATE = f"{ROOT_PATH}/pirate.txt"
    COT_MCTS = f"{ROOT_PATH}/cot.md"



# Read the 'system' message from a file
def read_system_message(filepath):
    try:
        with open(filepath, 'r') as file:
            return file.read().strip()
    except Exception as e:
        print(f"Error reading system message file: {e}")
        return "You are a helpful AI assistant."


import os
from langchain_core.messages import AIMessage, BaseMessage  
from langgraph.graph import END
from langchain_core.runnables import RunnableConfig
from langgraph.graph import MessagesState

from src.utils.state import State, StateManager
from src.utils.llm import LLMWrapper, ModelName

llm = LLMWrapper(model_name=ModelName.ANTHROPIC, api_key=os.getenv("ANTHROPIC_API_KEY"))

local_store = {}

# Define the function that calls the model
def call_chat_model(state: MessagesState, config: RunnableConfig) -> list[BaseMessage]:
    # Make sure that config is populated with the session id
    if "configurable" not in config or "session_id" not in config["configurable"]:
        raise ValueError(
            "Make sure that the config includes the following information: {'configurable': {'session_id': 'some_value'}}"
        )
    # Fetch the history of messages and append to it any new messages.
    state_manager = StateManager(config)
    messages = state_manager.collect_messages(local_store, state)
    ai_message = llm.invoke(messages)
    # Finally, update the chat message history to include
    # the new input message from the user together with the
    # repsonse from the model.
    state_manager.add_messages(local_store, messages + [ai_message])
    return {"messages": ai_message}

def error_node(state: State):
    return {"messages": [AIMessage(content="I'm sorry, I don't understand. Please try again.")]}

def should_run(state: State):
    return "call_chat_model" if 'please' in state["messages"][-1].content.lower() else END

def should_run(state: State):
    return "call_chat_model" if 'please' in state["messages"][-1].content.lower() else 'error'



from langchain_core.tools import tool

from langchain_community.tools import ShellTool
from langgraph.prebuilt import ToolNode
from langchain_core.documents import Document

from src.utils.vector import DEFAULT_VECTOR_STORE_PATH, VectorStore

@tool
def get_weather(location: str):
    """Call to get the current weather."""
    if location.lower() in ["sf", "san francisco"]:
        return "It's 60 degrees and foggy."
    else:
        return "It's 90 degrees and sunny."


@tool
def get_coolest_cities():
    """Get a list of coolest cities"""
    return "nyc, sf"

@tool
def shell_tool(commands: list[str]):
    """Run a shell commands. Commands is a list of strings, each representing a command to run. Avoid interactive commands."""
    shell_tool = ShellTool()
    output = shell_tool.run({"commands": commands})
    print(output)
    return output

@tool
def docker_shell_tool(commands: list[str]):
    """Run shell commands in a Docker container. Accepts multiple commands as a list of strings. 
    Each command is executed sequentially inside the specified container. Avoid interactive commands."""
    
    shell_tool = ShellTool()  # Initialize the ShellTool instance
    
    container_name = "ubuntu24"
    
    # Combine the docker execution command with each provided command
    docker_commands = [
        f'docker exec {container_name} bash -c "{command}"' for command in commands
    ]
    
    # Run the commands sequentially
    outputs = []
    for docker_command in docker_commands:
        output = shell_tool.run({"commands": [docker_command]})
        outputs.append(output)
        print(output)  # Optional: Log each command's output
    
    return outputs  # Return the output of all commands

@tool
def vector_store_query_tool(query: str):
    """Query the vector store. Search type can be 'mmr' or 'similarity'. Search kwargs is a dictionary of kwargs for the search type."""
    loaded_vector_store = VectorStore().load_vector_store()
    vector_store = VectorStore(loaded_vector_store)
    return vector_store.retrieve(query, "similarity", {"k": 10})

@tool
def vector_store_add_docs_tool(docs: list[Document]):
    """Add documents to the vector store.
    
    Example:

        .. code-block:: python

            from langchain_core.documents import Document

            document = Document(
                page_content="Hello, world!",
                metadata={"source": "https://example.com"}
            )
    """
    return VectorStore().add_docs(docs)

@tool
def vector_store_load_tool(path: str = DEFAULT_VECTOR_STORE_PATH):
    """Load the vector store from a file."""
    vector_store = VectorStore()
    return vector_store.load_vector_store(path)


tools = [
    get_weather, 
    get_coolest_cities, 
    shell_tool, 
    docker_shell_tool, 
    vector_store_query_tool, 
    vector_store_add_docs_tool,
    vector_store_load_tool
]
tool_node = ToolNode(tools)